import sys
import threading

import pyblish.api
import pyblish.lib
import pyblish.logic

from . import service as service_
from ..vendor.six.moves import (
    xmlrpc_server as xmlrpclib,
    socketserver
)


self = sys.modules[__name__]
self.current_server_thread = None
self.current_server = None
self.dispatch_wrapper = None


def default_wrapper(func, *args, **kwargs):
    return func(*args, **kwargs)


class RpcServer(socketserver.ThreadingMixIn, xmlrpclib.SimpleXMLRPCServer):
    """The Pyblish RPC Server

    Support multiple requests simultaneously. This is important,
    as we will still want to emit heartbeats during a potentially
    long process.

    About paths, the API is accessible from both the root /pyblish and
    suffixed with an exact version, such as /pyblish/v1. The root always
    references the latest version.

    """

    def __init__(self, path, *args, **kwargs):
        class VerifyingRequestHandler(xmlrpclib.SimpleXMLRPCRequestHandler):
            rpc_paths = ("/pyblish", path)

            def parse_request(this):
                if xmlrpclib.SimpleXMLRPCRequestHandler.parse_request(this):
                    if self.authenticate(this.headers):
                        return True
                else:
                    this.send_error(401, "Authentication failed")

                return False

        xmlrpclib.SimpleXMLRPCServer.__init__(
            self,
            requestHandler=VerifyingRequestHandler,
            *args,
            **kwargs)

    def authenticate(self, headers):
        """TODO(marcus): Implement basic authentication"""
        # basic, _, encoded = headers.get("Authorization").partition(" ")
        # assert basic == "Basic", "Only basic authentication supported"
        # username, _, password = base64.b64decode(encoded).partition(":")
        # assert username == "marcus"
        # assert password == "pass"
        return True

    def _dispatch(self, *args):
        wrapper = sys.modules[__name__].dispatch_wrapper or default_wrapper
        return wrapper(xmlrpclib.SimpleXMLRPCServer._dispatch,
                       self, *args)


def kill():
    """Shutdown a running server"""
    print("Shutting down..")
    return self.current_server.shutdown()


def _server(port, service):
    server = RpcServer(
        "/pyblish",
        ("127.0.0.1", port),
        allow_none=True,
        logRequests=False)

    server.register_function(kill)
    server.register_introspection_functions()
    server.register_instance(service, allow_dotted_names=True)

    self.current_server = server

    return server


def _serve(port, service=None):
    if service is None:
        service = service_.RpcService()

    server = _server(port, service)
    print("Listening on %s:%s" % server.server_address)
    server.serve_forever()


def start_production_server(port, service=None):
    """Run server with optimisations

    Arguments:
        port (int): Port at which to listen for incoming requests
        service (RpcService): Service responding to requests

    """

    return _serve(port, service)


def start_async_production_server(port, service=None):
    """Start a threaded version of production server

    Returns Thread object.

    """

    def worker():
        start_production_server(port, service)

    thread = threading.Thread(target=worker)
    thread.daemon = True
    thread.start()

    self.current_server_thread = thread

    return thread


def start_debug_server(port=6000, delay=0.5):
    """Start debug server

    This server uses a mocked up service to fake the actual
    behaviour and data of a generic host; incuding faked time
    it takes to perform a task.

    Arguments:
        port (int, optional): Port at which to listen for requests.
            Defaults to 6000.
        delay (float, optional): Simulate time taken to process requests.
            Defaults to 0.5 seconds.

    """

    pyblish.lib.setup_log("pyblish")
    service = service_.MockRpcService()
    return _serve(port, service)



import sys
import threading

from pyblish_qml import rpc

from PyQt5 import QtCore

self = sys.modules[__name__]

self.app = (
    QtCore.QCoreApplication.instance() or
    QtCore.QCoreApplication(sys.argv)
)

self.port = 50998
self.service = rpc.service.RpcService()
self.server = rpc.server._server(self.port, self.service)
self.thread = threading.Thread(target=self.server.serve_forever)
self.thread.daemon = True
self.thread.start()



import os
import sys
import time
import socket
import inspect
import traceback
import threading

# Python 2 and 3 compatibility
from .vendor.six.moves import xmlrpc_client as xmlrpclib

import pyblish.api

from . import rpc, settings


self = sys.modules[__name__]
self.ACTIVE_HOST_PORT = None
self.ACTIVE_PROXY = None
self.ACTIVE_HOST_NAME = None


def register_dispatch_wrapper(wrapper):
    """Register a dispatch wrapper for servers

    The wrapper must have this exact signature:
        (func, *args, **kwargs)

    """

    signature = inspect.getargspec(wrapper)
    if any([len(signature.args) != 1,
            signature.varargs is None,
            signature.keywords is None]):
        raise TypeError("Wrapper signature mismatch")

    rpc.server.dispatch_wrapper = wrapper


def deregister_dispatch_wrapper():
    rpc.server.dispatch_wrapper = None


def dispatch_wrapper():
    return rpc.server.dispatch_wrapper


def install(initial_port=9001):
    """Perform first time install

    Attributes:
        initial_port (int, optional): Port from which to start
            looking for available ports, defaults to 9001

    """

    if self.ACTIVE_HOST_PORT is not None:
        uninstall()

    install_host()
    install_callbacks()

    try:
        # In case QML is live and well, ask it
        # for the next available initial_port number.
        self.ACTIVE_PROXY = proxy()
        self.ACTIVE_HOST_PORT = self.ACTIVE_PROXY.find_available_port(
            initial_port)

    except (socket.timeout, socket.error):
        raise socket.error(
            "Is Pyblish QML running?\n\n"

            "This GUI requires a separate, standalone process to be running\n"
            "in the background in order to make use of it from within {host}\n"
            "Make sure to run python -m pyblish_qml prior to showing it from\n"
            "from {host}.\n\n"

            "See the documentation for further information.\n"
            "- https://github.com/pyblish/pyblish-qml\n\n#"
            .format(
                host=self.ACTIVE_HOST_NAME or "a host")
        )

    os.environ["PYBLISH_CLIENT_PORT"] = str(self.ACTIVE_HOST_PORT)

    try:
        _serve()

    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        message = "".join(traceback.format_exception(
            exc_type, exc_value, exc_traceback))
        sys.stderr.write(message)
        sys.stderr.write("Setup failed..\n")


def uninstall():
    """Kill server and clean up traces of Pyblish QML in this process"""
    if self.ACTIVE_PROXY is None:
        return

    del self.ACTIVE_PROXY

    self.ACTIVE_HOST_PORT = None
    self.ACTIVE_PROXY = None

    uninstall_callbacks()
    uninstall_server()

    sys.stdout.write("Pyblish QML shutdown successful.\n")


def show():
    """Attempt to show GUI

    Requires install() to have been run first, and
    a live instance of Pyblish QML in the background.

    """

    if not self.ACTIVE_HOST_PORT:
        sys.stdout.write("Running install() for the first time.\n")
        install()

    if self.ACTIVE_PROXY is None:
        self.ACTIVE_PROXY = proxy()

    try:
        self.ACTIVE_PROXY.show(
            self.ACTIVE_HOST_PORT,
            settings.to_dict()
        )

    except (socket.error, socket.timeout):
        raise socket.error(
            "Communication problem.\n"
            "\n"
            "Hello? Is it me you're looking for?\n"
            "\n"
            "This problem typically occurs when Pyblish QML \n"
            "was running, but isn't anymore.\n"
            "\n"
            "To start it again, go to any terminal any type:\n"
            "$ python -m pyblish_qml"
        )


def install_callbacks():
    pyblish.api.register_callback("instanceToggled", _toggle_instance)


def uninstall_callbacks():
    pyblish.api.deregister_callback("instanceToggled", _toggle_instance)


def uninstall_server():
    rpc.server.kill()


def _serve():
    """Start XML-RPC server

    This is the entrypoint towards which the remote Pyblish QML instance
    communicates.

    """

    assert self.ACTIVE_HOST_PORT

    def server():
        """Provide QML with a friend to speak with"""
        self.server = rpc.server.start_production_server(self.ACTIVE_HOST_PORT)

    def heartbeat_emitter():
        """Let QML know we're still here"""
        p = proxy()

        while True:
            try:
                p.heartbeat(self.ACTIVE_HOST_PORT)
                time.sleep(1)
            except (socket.error, socket.timeout):
                pass

    for worker in (server, heartbeat_emitter):
        t = threading.Thread(target=worker, name=worker.__name__)
        t.daemon = True
        t.start()

    sys.stdout.write("Server running @ %i\n" % self.ACTIVE_HOST_PORT)

    return self.ACTIVE_HOST_PORT


def _toggle_instance(instance, new_value, old_value):
    """Alter instance upon visually toggling instance"""
    instance.data["publish"] = new_value


def proxy(timeout=5):
    """Return proxy at default location of Pyblish QML"""
    return xmlrpclib.ServerProxy(
        "http://127.0.0.1:9090",
        allow_none=True)


def install_host():
    """Install required components into supported hosts

    An unsupported host will still run, but may encounter issues,
    especially with threading.

    """

    for install in (_install_maya,
                    _install_houdini,
                    _install_nuke):
        try:
            install()
        except ImportError:
            pass
        else:
            break


def _install_maya():
    """Helper function to Autodesk Maya support"""
    from maya import utils

    def threaded_wrapper(func, *args, **kwargs):
        return utils.executeInMainThreadWithResult(
            func, *args, **kwargs)

    sys.stdout.write("Setting up Pyblish QML in Maya\n")
    register_dispatch_wrapper(threaded_wrapper)

    # Configure GUI
    settings.ContextLabel = "Maya"
    settings.WindowTitle = "Pyblish (Maya)"

    self.ACTIVE_HOST_NAME = "Maya"


def _install_houdini():
    """Helper function to SideFx Houdini support"""
    import hdefereval

    def threaded_wrapper(func, *args, **kwargs):
        return hdefereval.executeInMainThreadWithResult(
            func, *args, **kwargs)

    sys.stdout.write("Setting up Pyblish QML in Houdini\n")
    register_dispatch_wrapper(threaded_wrapper)

    settings.ContextLabel = "Houdini"
    settings.WindowTitle = "Pyblish (Houdini)"

    self.ACTIVE_HOST_NAME = "Houdini"


def _install_nuke():
    """Helper function to The Foundry Nuke support"""
    import nuke

    if "--hiero" in nuke.rawArgs or "--studio" in nuke.rawArgs:
        raise ImportError

    def threaded_wrapper(func, *args, **kwargs):
        return nuke.executeInMainThreadWithResult(
            func, args, kwargs)

    sys.stdout.write("Setting up Pyblish QML in Nuke\n")
    register_dispatch_wrapper(threaded_wrapper)

    settings.ContextLabel = "Nuke"
    settings.WindowTitle = "Pyblish (Nuke)"

    self.ACTIVE_HOST_NAME = "Nuke"


def _install_hiero():
    """Helper function to The Foundry Hiero support"""
    import hiero
    import nuke

    if "--hiero" not in nuke.rawArgs:
        raise ImportError

    def threaded_wrapper(func, *args, **kwargs):
        return hiero.core.executeInMainThreadWithResult(
            func, args, kwargs)

    sys.stdout.write("Setting up Pyblish QML in Hiero\n")
    register_dispatch_wrapper(threaded_wrapper)

    settings.ContextLabel = "Hiero"
    settings.WindowTitle = "Pyblish (Hiero)"

    self.ACTIVE_HOST_NAME = "Hiero"



from PyQt5 import QtTest

# Vendor libraries
from nose.tools import (
    assert_in,
    assert_true,
    assert_equals,
    with_setup
)

from pyblish_qml import control
import pyblish.api

from . import port, lib


def check_present(name, model):
    assert_in(name, [i.name for i in model.items])


def reset(controller=None):
    if controller is None:
        controller = control.Controller()
        controller.on_client_changed(port)

    ready = QtTest.QSignalSpy(controller.ready)

    assert_true(ready.wait(1000))

    count = len(ready)
    controller.reset()

    ready.wait(1000)
    assert_equals(len(ready), count + 1)
    assert_true("ready" in controller.states)

    return controller


def publish(controller=None):
    """Publish current state

    Argument:
        plugins (list): List of plug-ins to include
        plugin (str): Name of plug-in to repair

    """

    if controller is None:
        controller = control.Controller()
        controller.on_client_changed(port)

    finished = QtTest.QSignalSpy(controller.finished)

    count = len(finished)
    controller.publish()

    finished.wait(1000)
    assert_equals(len(finished), count + 1)
    assert_true("finished" in controller.states)

    return controller


def validate(controller=None):
    """Publish current state

    Argument:
        plugins (list): List of plug-ins to include
        plugin (str): Name of plug-in to repair

    """

    if controller is None:
        controller = control.Controller()
        controller.on_client_changed(port)

    finished = QtTest.QSignalSpy(controller.finished)

    count = len(finished)
    controller.validate()

    finished.wait(1000)
    assert_equals(len(finished), count + 1)
    assert_true("finished" in controller.states)

    return controller


@with_setup(lib.clean)
def test_reset():
    """Reset works"""

    data = {"#": 0, "instances": list()}

    class MyCollector(pyblish.api.Collector):
        def process(self, context):
            instance = context.create_instance("MyInstance")
            instance.data["family"] = "myFamily"
            data["#"] += 1
            data["instances"].append(instance)

    pyblish.api.register_plugin(MyCollector)

    c = reset()

    # At this point, the item-model is populated with
    # a number of instances.
    check_present("MyCollector", c.item_model)
    check_present(data["instances"][0].name, c.item_model)
    assert_equals(data["#"], 1)


@with_setup(lib.clean)
def test_publish():
    """Publishing works"""

    count = {"#": 0}

    class Collector(pyblish.api.Collector):
        def process(self, context):
            instance = context.create_instance("MyInstance")
            instance.data["family"] = "myFamily"
            count["#"] += 1

    class Validator(pyblish.api.Validator):
        def process(self, instance):
            count["#"] += 10

    class Extractor(pyblish.api.Extractor):
        def process(self, instance):
            count["#"] += 100

    plugins = [Collector, Validator, Extractor]

    for plugin in plugins:
        pyblish.api.register_plugin(plugin)

    c = reset()

    check_present("Collector", c.item_model)
    check_present("MyInstance", c.item_model)
    assert_equals(count["#"], 1)

    # At this point, the item-model is populated with
    # a number of instances.

    publish(c)

    names = [p.__name__ for p in plugins]
    inmodel = [p.name for p in c.item_model.plugins]

    assert_true(all(n in inmodel for n in names))
    assert_equals(len(c.host.context()), 1)
    assert_equals(count["#"], 111)


@with_setup(lib.clean)
def test_publish_only_toggled():
    """Only toggled items are published"""

    count = {"#": 0}

    class MyCollector(pyblish.api.Collector):
        def process(self, context):
            instance = context.create_instance("MyInstance")
            instance.data["family"] = "myFamily"
            count["#"] += 1

    class MyValidator(pyblish.api.Validator):
        def process(self, instance):
            count["#"] += 10

    class MyExtractor(pyblish.api.Extractor):
        def process(self, instance):
            count["#"] += 100

    plugins = [MyCollector, MyValidator, MyExtractor]

    for plugin in plugins:
        pyblish.api.register_plugin(plugin)

    c = reset()

    check_present("MyCollector", c.item_model)
    check_present("MyInstance", c.item_model)

    c.item_model.plugins[MyValidator.id].isToggled = False

    publish(c)

    assert_equals(count["#"], 101)


@with_setup(lib.clean)
def test_cooperative_collection():
    """Cooperative collection works

    A collector should be able to run such that the following
    collector "sees" the newly created instance so as to
    query and/or modify it.

    """

    history = []
    count = {"#": 0}

    class CollectorA(pyblish.api.Collector):
        order = 0.0

        def process(self, context):
            context.create_instance("myInstance1")
            history.append(type(self).__name__)
            count["#"] += 1

    class CollectorB(pyblish.api.Collector):
        order = 0.1

        def process(self, context):
            assert "myInstance1" in [i.data["name"] for i in context]

            # This should run
            history.append(type(self).__name__)
            count["#"] += 10

            context.create_instance("myInstance2")

    pyblish.api.register_plugin(CollectorA)
    pyblish.api.register_plugin(CollectorB)

    c = reset()

    check_present("CollectorA", c.item_model)
    check_present("CollectorB", c.item_model)
    check_present("myInstance1", c.item_model)
    check_present("myInstance2", c.item_model)

    assert count["#"] == 11, count
    assert history == ["CollectorA", "CollectorB"]


@with_setup(lib.clean)
def test_argumentless_plugin():
    """Implicit plug-ins without arguments should still run"""
    count = {"#": 0}

    class MyPlugin(pyblish.api.Validator):
        def process(self):
            count["#"] += 1

    pyblish.api.register_plugin(MyPlugin)

    c = reset()
    publish(c)

    check_present("MyPlugin", c.item_model)

    assert count["#"] == 1


@with_setup(lib.clean)
def test_cooperative_collection2():
    """Cooperative collection works with InstancePlugin"""

    count = {"#": 0}
    history = []

    class CollectorA(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            history.append(type(self).__name__)
            context.create_instance("MyInstance")
            count["#"] += 1

    class CollectorB(pyblish.api.InstancePlugin):
        order = pyblish.api.CollectorOrder + 0.1

        def process(self, instance):
            history.append(type(self).__name__)
            count["#"] += 10

    pyblish.api.register_plugin(CollectorA)
    pyblish.api.register_plugin(CollectorB)

    c = reset()

    check_present("CollectorA", c.item_model)
    check_present("CollectorB", c.item_model)
    check_present("MyInstance", c.item_model)

    assert count["#"] == 11, count
    assert history == ["CollectorA", "CollectorB"]


@with_setup(lib.clean)
def test_published_event():
    """published is emitted upon finished publish"""

    count = {"#": 0}

    def on_published(context):
        assert isinstance(context, pyblish.api.Context)
        count["#"] += 1

    pyblish.api.register_callback("published", on_published)

    c = reset()
    publish(c)

    assert count["#"] == 1, count


@with_setup(lib.clean)
def test_validated_event():
    """validated is emitted upon finished validation"""

    count = {"#": 0}

    def on_validated(context):
        assert isinstance(context, pyblish.api.Context)
        count["#"] += 1

    pyblish.api.register_callback("validated", on_validated)

    c = reset()
    validate(c)

    assert count["#"] == 1, count


@with_setup(lib.clean)
def test_gui_vs_host_order():
    """Host properly reflects order of instances in GUI"""

    instances = [
        "instance4",
        "instance3",
        "instance2",
        "instance1"
    ]

    class Collector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            for name in instances:
                context.create_instance(name)

    class SortByName(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder + 0.1

        def process(self, context):
            context[:] = sorted(context, key=lambda i: i.name)

    pyblish.api.register_plugin(Collector)

    # Test baseline
    c = reset()

    gui_instances = list(
        i.name for i in c.item_model.instances
        if i.name != "Context"
    )

    # GUI should contain the original order
    assert gui_instances == instances, "QML has got a different order"

    host_instances = list(i.name for i in c.host.context())

    # Host should also contain the original order
    assert host_instances == instances, "Host has got a different order"

    # Sort context retrospectively
    pyblish.api.register_plugin(SortByName)
    c = reset()

    gui_instances = list(
        i.name for i in c.item_model.instances
        if i.name != "Context"
    )

    host_instances = list(i.name for i in c.host.context())

    # GUI should reflect sorted order
    print("gui_instances: %s" % gui_instances)
    assert gui_instances != instances, "Sorting did not happen in QML"

    print("host_instances: %s" % host_instances)
    assert host_instances != instances, "Sorting did not happen in host"

    assert gui_instances == host_instances, "QML != Host"


@with_setup(lib.clean)
def test_toggle_compatibility():
    """toggle instance updates compatibility correctly"""

    data = {"instance": None}

    class Collector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            instance = context.create_instance("A")
            instance.data["family"] = "FamilyA"
            data['instance'] = instance

    class Validate(pyblish.api.InstancePlugin):
        """A dummy validator"""
        order = pyblish.api.ValidatorOrder
        families = ["FamilyA"]

        def process(self, instance):
            pass

    pyblish.api.register_plugin(Collector)
    pyblish.api.register_plugin(Validate)

    c = reset()

    def has_enabled_validator(c):
        """Return whether Validate validator is enabled/compatible"""

        rows = c.plugin_proxy.rowCount()
        for row in range(rows):
            item = c.plugin_proxy.item(row)
            if item.name == 'Validate':
                return True

        return False

    item = c.item_model.instances[data['instance'].id]
    index = c.item_model.items.index(item)

    # Default state (enabled)
    assert has_enabled_validator(c), "Not enabled before"

    # toggle off
    c.toggleInstance(index)
    assert not has_enabled_validator(c), "Not disabled"

    # toggle back on
    c.toggleInstance(index)
    assert has_enabled_validator(c), "Not enabled after"


@with_setup(lib.clean)
def test_action_on_failed():
    """Actions on failed are exclusive to plug-ins that have failed"""

    class SelectMany(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            for name in ("A", "B", "C"):
                context.create_instance(name)

    class ActionOnFailed(pyblish.api.Action):
        on = "failed"

    class ValidateFail(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder
        actions = [ActionOnFailed]

        def process(self, instance):
            assert False

    class ValidateSuccess(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder
        actions = [ActionOnFailed]

        def process(self, instance):
            assert True

    pyblish.api.register_plugin(SelectMany)
    pyblish.api.register_plugin(ValidateFail)
    pyblish.api.register_plugin(ValidateSuccess)

    c = reset()

    validate_fail_index = c.item_model.items.index(
        c.item_model.plugins[ValidateFail.id])
    validate_success_index = c.item_model.items.index(
        c.item_model.plugins[ValidateSuccess.id])

    validate_fail_actions = c.getPluginActions(validate_fail_index)

    assert not validate_fail_actions, (
        "ValidateFail should not have produced any actions")

    validate(c)

    validate_fail_actions = c.getPluginActions(validate_fail_index)

    assert len(validate_fail_actions) == 1, (
        "ValidateFail should have had an action")
    assert validate_fail_actions[0]["id"] == ActionOnFailed.id, (
        "ValidateFail had an unknown action")

    validate_success_actions = c.getPluginActions(validate_success_index)
    assert len(validate_success_actions) == 0, (
        "ValidateSuccess should not have had an action")


@with_setup(lib.clean)
def test_action_not_processed():
    """Actions on not processed """

    class SelectMany(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder

        def process(self, context):
            for name in ("A", "B", "C"):
                context.create_instance(name)

    class ActionOnNotProcessed(pyblish.api.Action):
        on = "notProcessed"

    class ValidateAction(pyblish.api.InstancePlugin):
        order = pyblish.api.ValidatorOrder
        actions = [ActionOnNotProcessed]

        def process(self, instance):
            assert True

    pyblish.api.register_plugin(SelectMany)
    pyblish.api.register_plugin(ValidateAction)

    c = reset()

    validate_index = c.item_model.items.index(
        c.item_model.plugins[ValidateAction.id])

    validate_actions = c.getPluginActions(validate_index)

    assert len(validate_actions) == 1, (
        "ValidateAction should have 1 action")

    validate(c)

    validate_actions = c.getPluginActions(validate_index)

    assert len(validate_actions) == 0, (
        "ValidateAction should not have had an action")


def test_inactive_collector():
    """An inactive collector should not run"""

    count = {"#": 0}

    class MyInactiveCollector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder
        active = False

        def process(self, context):
            print("Ran %s" % type(self))
            count["#"] += 1

    class MyActiveCollector(pyblish.api.ContextPlugin):
        order = pyblish.api.CollectorOrder
        active = True

        def process(self, context):
            print("Ran %s" % type(self))
            count["#"] += 10

    pyblish.api.register_plugin(MyInactiveCollector)
    pyblish.api.register_plugin(MyActiveCollector)

    reset()

    assert count["#"] == 10



"""Client communication library

This library is responsible for intercepting and processing
both incoming and outgoing communication. Incoming communication
is parsed into "Object Proxys" (see below) whereas outgoing
communication is serialised back into its original JSON.

"""

import socket

from ..vendor.six.moves import xmlrpc_client as xmlrpclib

# Dependencies
import pyblish.api
import pyblish.plugin


class Proxy(object):
    """Wrap ServerProxy with logic and object proxies

    The proxy mirrors the remote interface to provide an
    as-similar experience as possible.

    """

    _instance = None

    def __getattr__(self, attr):
        """Any call not overloaded, simply pass it on"""
        return getattr(self._proxy, attr)

    def __init__(self, port, user=None, password=None):
        self.cached_context = list()
        self.cached_discover = list()

        transport = xmlrpclib.Transport()

        self._proxy = xmlrpclib.ServerProxy(
            "http://{auth}127.0.0.1:{port}/pyblish".format(
                port=port,
                auth=("{user}:{pwd}@".format(
                    user=user, pwd=password)
                ) if user else ""),
            allow_none=True,
            transport=transport)

    def test(self, **vars):
        """Vars can only be passed as a non-keyword argument"""
        return self._proxy.test(vars)

    def ping(self):
        """Convert Fault to True/False"""
        try:
            self._proxy.ping()
        except (socket.timeout, socket.error):
            return False
        return True

    def process(self, plugin, context, instance=None, action=None):
        """Transmit a `process` request to host

        Arguments:
            plugin (PluginProxy): Plug-in to process
            context (ContextProxy): Filtered context
            instance (InstanceProxy, optional): Instance to process
            action (str, optional): Action to process

        """

        plugin = plugin.to_json()
        instance = instance.to_json() if instance is not None else None
        return self._proxy.process(plugin, instance, action)

    def repair(self, plugin, context, instance=None):
        plugin = plugin.to_json()
        instance = instance.to_json() if instance is not None else None
        return self._proxy.repair(plugin, instance)

    def context(self):
        self.cached_context = ContextProxy.from_json(self._proxy.context())
        return self.cached_context

    def discover(self):
        self.cached_discover[:] = list()
        for plugin in self._proxy.discover():
            self.cached_discover.append(PluginProxy.from_json(plugin))

        return self.cached_discover

    def emit(self, signal, **kwargs):
        self._proxy.emit(signal, kwargs)


# Object Proxies


class ContextProxy(pyblish.api.Context):
    """Context Proxy

    Given a JSON-representation of a Context, emulate its interface.

    """

    def create_instance(self, name, **kwargs):
        instance = InstanceProxy(name, parent=self)
        instance.data.update(kwargs)
        return instance

    @classmethod
    def from_json(cls, context):
        self = cls()
        self._id = context["id"]
        self._data = context["data"]
        self[:] = list(InstanceProxy.from_json(i)
                       for i in context["children"])

        # Attach metadata
        self._data["pyblishClientVersion"] = pyblish.api.version

        return self

    def to_json(self):
        return {
            "name": self.name,
            "id": self.id,
            "data": dict(self.data),  # Convert pyblish.plugin._Dict object
            "children": list(self),
        }


class InstanceProxy(pyblish.api.Instance):
    """Instance Proxy

    Given a JSON-representation of an Instance, emulate its interface.

    """

    @classmethod
    def from_json(cls, instance):
        self = cls(instance["name"])
        self._id = instance["id"]
        self._data = instance["data"]
        self[:] = instance["children"]

        return self

    def to_json(self):
        return {
            "name": self.name,
            "id": self.id,
            "data": dict(self.data),
            "children": list(self),
        }


class PluginProxy(object):
    """Plug-in Proxy

    Given a JSON-representation of an Plug-in, emulate its interface.

    """

    @classmethod
    def from_json(cls, plugin):
        """Build PluginProxy object from incoming dictionary

        Emulate a plug-in by providing access to attributes
        in the same way they are accessed using the remote object.
        This allows for it to be used by members of :mod:`pyblish.logic`.

        """

        process = None
        repair = None

        name = plugin["name"] + "Proxy"
        cls = type(name, (cls,), plugin)

        # Emulate function
        for name in ("process", "repair"):
            args = ", ".join(plugin["process"]["args"])
            func = "def {name}({args}): pass".format(name=name,
                                                     args=args)
            exec(func)

        cls.process = process
        cls.repair = repair

        cls.__orig__ = plugin

        return cls

    @classmethod
    def to_json(cls):
        return cls.__orig__.copy()



from .version import (
    version,
    version_info,
    __version__
)

from .host import (
    register_dispatch_wrapper,
    deregister_dispatch_wrapper,
    dispatch_wrapper,
    uninstall,
    install,
    show,
)

__all__ = [
    "__version__",
    "version",
    "version_info",
    "register_dispatch_wrapper",
    "deregister_dispatch_wrapper",
    "dispatch_wrapper",
    "install",
    "uninstall",
    "show",
]



"""Pyblish QML command-line interface"""

import sys
import socket
import argparse

from . import app, host


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("--demo", action="store_true",
                        help="Illustrate features of the GUI")
    parser.add_argument("--debug", action="store_true",
                        help="Development mode")
    parser.add_argument("--validate", action="store_true",
                        help="Check system prior to launching")
    parser.add_argument("--kill", action="store_true",
                        help="Kill running QML client")
    kwargs = parser.parse_args()

    # If a QML is already running, re-use this.
    proxy = host.proxy(timeout=5)

    try:
        proxy.ping()

    except (socket.timeout, socket.error):
        return app.main(debug=kwargs.debug or kwargs.demo,
                        validate=kwargs.validate)

    else:
        if kwargs.kill:
            proxy.kill()
            print("Successfully killed client.")


sys.exit(cli())



"""Application entry-point"""

# Standard library
import os
import sys
import time
import threading

# Dependencies
from PyQt5 import QtCore, QtGui, QtQuick, QtTest

# Local libraries
from . import util, compat, server, control, rpc, settings

MODULE_DIR = os.path.dirname(__file__)
QML_IMPORT_DIR = os.path.join(MODULE_DIR, "qml")
APP_PATH = os.path.join(MODULE_DIR, "qml", "main.qml")
ICON_PATH = os.path.join(MODULE_DIR, "icon.ico")


class Window(QtQuick.QQuickView):
    """Main application window"""

    def __init__(self, parent=None):
        super(Window, self).__init__(None)
        self.parent = parent

        self.setTitle(settings.WindowTitle)
        self.setResizeMode(self.SizeRootObjectToView)

        self.resize(*settings.WindowSize)
        self.setMinimumSize(QtCore.QSize(430, 300))

    def event(self, event):
        """Allow GUI to be closed upon holding Shift"""
        if event.type() == QtCore.QEvent.Close:
            modifiers = self.parent.queryKeyboardModifiers()
            shift_pressed = QtCore.Qt.ShiftModifier & modifiers

            if shift_pressed or hasattr(self.parent, "__debugging__"):
                event.accept()

            elif "publishing" in self.parent.controller.states:
                event.ignore()

            else:
                event.ignore()
                self.parent.hide()

        return super(Window, self).event(event)


class Application(QtGui.QGuiApplication):
    """Pyblish QML wrapper around QGuiApplication

    Provides production and debug launchers along with controller
    initialisation and orchestration.

    """

    show_signal = QtCore.pyqtSignal(int, QtCore.QVariant)
    hide_signal = QtCore.pyqtSignal()
    quit_signal = QtCore.pyqtSignal()
    app_message = QtCore.pyqtSignal(str)
    standalone_signal = QtCore.pyqtSignal()

    def __init__(self, source):
        super(Application, self).__init__(sys.argv)

        self.setWindowIcon(QtGui.QIcon(ICON_PATH))

        window = Window(self)
        window.statusChanged.connect(self.on_status_changed)

        engine = window.engine()
        engine.addImportPath(QML_IMPORT_DIR)

        controller = control.Controller()

        context = engine.rootContext()
        context.setContextProperty("app", controller)

        self.window = window
        self.engine = engine
        self.controller = controller
        self.clients = dict()
        self.current_client = None

        self.show_signal.connect(self.show)
        self.hide_signal.connect(self.hide)
        self.quit_signal.connect(self.quit)
        self.standalone_signal.connect(self.on_standalone)

        window.setSource(QtCore.QUrl.fromLocalFile(source))

    def on_status_changed(self, status):
        if status == QtQuick.QQuickView.Error:
            self.quit()

    def register_heartbeat(self, port):
        if port not in self.clients:
            self.register_client(port)
        self.clients[port]["lastSeen"] = time.time()

    def register_client(self, port):
        self.current_client = port
        self.clients[port] = {
            "lastSeen": time.time()
        }

    def deregister_client(self, port):
        self.clients.pop(port)

    def show(self, port, client_settings=None):
        """Display GUI

        Once the QML interface has been loaded, use this
        to display it.

        Arguments:
            port (int): Client asking to show GUI.
            client_settings (dict, optional): Visual settings, see settings.py

        """

        window = self.window

        if client_settings:
            # Apply client-side settings
            settings.from_dict(client_settings)
            window.setWidth(client_settings["WindowSize"][0])
            window.setHeight(client_settings["WindowSize"][1])
            window.setTitle(client_settings["WindowTitle"])

        print("Settings:")
        for key, value in client_settings.items():
            print("  %s = %s" % (key, value))

        previously_hidden = not window.isVisible()

        window.requestActivate()
        window.showNormal()

        new_client = False
        if settings.current_host_port() != port:
            new_client = True
            settings.set_current_host_port(port)

        if os.name == "nt":
            # Work-around for window appearing behind
            # other windows upon being shown once hidden.
            previous_flags = window.flags()
            window.setFlags(previous_flags | QtCore.Qt.WindowStaysOnTopHint)
            window.setFlags(previous_flags)

        if previously_hidden or new_client:
            # Give statemachine enough time to boot up
            if not any(state in self.controller.states
                       for state in ["ready", "finished"]):
                util.timer("ready")

                ready = QtTest.QSignalSpy(self.controller.ready)

                count = len(ready)
                ready.wait(1000)
                if len(ready) != count + 1:
                    print("Warning: Could not enter ready state")

                util.timer_end("ready", "Awaited statemachine for %.2f ms")

            self.controller.show.emit()
            self.controller.reset()

    def hide(self):
        """Hide GUI

        Process remains active and may be shown
        via a call to `show()`

        """

        self.window.hide()

    def on_standalone(self):
        """The process is running solo"""
        self.app_message.emit("Running standalone")
        if not hasattr(self, "__debugging__"):
            print("Quitting due to loneliness")
            time.sleep(1)
            self.quit()

    def listen(self):
        """Listen on incoming messages from host

        Usage:
            >> from pyblish_qml import client
            >> proxy = client.proxy()
            >> proxy.show()   # Show a hidden GUI
            >> proxy.close()  # Close GUI permanently
            >> proxy.kill()   # Close GUI forcefully (careful)

        """

        kwargs = {
            "port": 9090,
            "service": server.QmlApi(self),
        }

        t = threading.Thread(target=server._serve, kwargs=kwargs)
        t.daemon = True
        t.start()


def main(source=None, debug=False, validate=True):
    """Start the Qt-runtime and show the window

    Arguments:
        source (str): QML entry-point
        debug (bool, optional): Run in debug-mode. Defaults to False
        validate (bool, optional): Whether the environment should be validated
            prior to launching. Defaults to True

    """

    # Initialise OS compatiblity
    compat.main()

    print("Starting Pyblish..")
    util.timer("application")

    # debug mode
    if debug:
        port = 6000

        app = Application(source or APP_PATH)
        app.listen()
        app.__debugging__ = True

        print("Starting in debug-mode")
        print("Looking for server..")
        proxy = rpc.client.Proxy(port)

        if not proxy.ping():
            os.environ["PYBLISH_CLIENT_PORT"] = str(port)

            print("No existing server found, creating..")
            thread = threading.Thread(
                target=rpc.server.start_debug_server,
                kwargs={"port": port})
            thread.daemon = True
            thread.start()

            print("Debug server created successfully.")
            print("Running mocked RPC server @ 127.0.0.1:%s" % port)

        app.show_signal.emit(port, {
            "ContextLabel": "World",
            "WindowTitle": "Pyblish (DEBUG)",
            "WindowSize": (430, 600)
        })

    else:
        app = Application(source or APP_PATH)
        app.listen()

    util.timer_end("application", "Spent %.2f ms creating the application")

    return app.exec_()


if __name__ == "__main__":
    main(debug=True,
         validate=False)



"""Pyblish QML RPC server

Attributes:
    first_port (int): Port at which to start distributing
        available ports to clients who ask for one.

"""

import os

from .vendor.six.moves import xmlrpc_server as xmlrpclib

first_port = 9001


class QmlApi(object):
    def __init__(self, app):
        self.app = app
        self.ctrl = CtrlApi(app.controller)

    def show(self, port, settings=None):
        """Show the GUI

        Arguments:
            port (int): Port with which to communicate with client
            settings (optional, dict): Client settings

        """

        self.app.show_signal.emit(port, settings)
        return True

    def hide(self):
        """Hide the GUI"""
        self.app.hide_signal.emit()
        return True

    def quit(self):
        """Ask the GUI to quit"""
        self.app.quit_signal.emit()
        return True

    def kill(self):
        """Forcefully destroy the process, this does not return"""
        os._exit(1)

    def heartbeat(self, port):
        """Tell QML that someone is listening at `port`"""
        self.app.register_heartbeat(port)

    def ping(self):
        """Used to check connectivity"""
        return {
            "message": "Hello, whomever you are"
        }

    def find_available_port(self, start=first_port):
        """Return the next available port at which a client may listen

        If module "psutil" is available, this also takes into
        account any externally used ports such that no occupied
        port is accidentally used. This is generally recommended.

        Arguments:
            start (int, optional): Port from which to start
                looking, defaults to 6001

        """

        print("Finding available port..")
        occupied_ports = list(self.app.clients)

        try:
            import psutil
            occupied_ports += list(
                c.laddr[-1] for c in psutil.net_connections())
        except ImportError:
            pass

        available = start
        while available in occupied_ports:
            available += 1

        print("Distributing new port %i" % available)
        return available


class CtrlApi(object):
    def __init__(self, ctrl):
        self.ctrl = ctrl


def _server(port, service):
    server = xmlrpclib.SimpleXMLRPCServer(
        ("127.0.0.1", port),
        allow_none=True,
        logRequests=False)

    server.register_introspection_functions()
    server.register_instance(service)

    return server


def _serve(port, service):
    server = _server(port, service)
    print("Listening on %s:%s" % server.server_address)
    return server.serve_forever()



import sys

# User-editable settings

ContextLabel = "Context"
WindowTitle = "Pyblish"
WindowSize = (430, 600)
WindowPosition = (100, 100)
HeartbeatInterval = 60

# Implementation details below.

self = sys.modules[__name__]
self._callbacks = dict()
self._current_gui_port = 0
self._current_host_port = 0
self._port_changed_callbacks = list()


def from_dict(settings):
    """Apply settings from dictionary

    Arguments:
        settings (dict): Settings in the form of a dictionary

    """

    assert isinstance(settings, dict), "`settings` must be of type dict"
    for key, value in settings.items():
        setattr(self, key, value)


def to_dict():
    """Return dictionary of settings"""
    return dict((k, getattr(self, k)) for k in {
        "ContextLabel",
        "WindowTitle",
        "WindowSize",
        "WindowPosition",
        "HeartbeatInterval",
    })


def current_host_port():
    """Return port through which QML is currently communicating"""
    return self._current_host_port


def set_current_host_port(port):
    """Set the port with which to communicate"""
    self._current_host_port = port
    for func in self._port_changed_callbacks:
        func(port)


def register_port_changed_callback(func):
    """When the current port changes, trigger the supplied callable

    Arguments:
        func (callable): Callable to run upon a new client registering
            interest in displaying the GUI.

    """

    self._port_changed_callbacks.append(func)
